/*
  Copyright 2019 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.auth

import com.kakao.sdk.auth.model.CertTokenInfo
import com.kakao.sdk.auth.model.OAuthToken
import com.kakao.sdk.common.KakaoSdk
import com.kakao.sdk.common.model.*

class AuthApiClient (
        private  val manager: AuthApiManager = AuthApiManager.instance,
    /** @suppress */ val tokenManagerProvider: TokenManagerProvider = TokenManagerProvider.instance )
{
    /**
     * 사용자가 앞서 로그인을 통해 토큰을 발급 받은 상태인지 확인합니다.
     * 주의: 기존 토큰 존재 여부를 확인하는 기능으로, 사용자가 현재도 로그인 상태임을 보장하지 않습니다.
     */
    fun hasToken() : Boolean = manager.hasToken()


    /**
     * 사용자 인증코드를 이용하여 신규 토큰 발급을 요청합니다.
     *
     * @param code 사용자 인증 코드
     * @param codeVerifier 사용자 인증 코드 verifier
     * @param callback 발급 받은 [OAuthToken] 반환.
     */
    fun issueAccessToken(
        code: String,
        codeVerifier: String? = null,
        callback: (token: OAuthToken?, error: Throwable?) -> Unit
    ) = manager.issueAccessToken(code, codeVerifier, callback)


    /**
     * @suppress
     * Internal Only
     */
    fun issueAccessTokenWithCert(
        code: String,
        codeVerifier: String? = null,
        callback: (certTokenInfo: CertTokenInfo?, error: Throwable?) -> Unit
    ) = manager.issueAccessTokenWithCert(code, codeVerifier, callback)


    /**
     * 기존 토큰을 갱신합니다
     *
     * @param oldToken 기존 토큰
     * @param callback 발급 받은 [OAuthToken] 반환.
     */
    @JvmOverloads
    fun refreshAccessToken(
        oldToken: OAuthToken = tokenManagerProvider.manager.getToken() ?: throw ClientError(ClientErrorCause.TokenNotFound, "Refresh token not found. You must login first."),
        callback: (token: OAuthToken?, error: Throwable?) -> Unit
    ) = manager.refreshAccessToken(oldToken, callback)


    /**
     * @suppress
     */
    fun agt(
        callback: (agt: String?, error: Throwable?) -> Unit
    ) = manager.agt(callback)

    companion object {
        @JvmStatic
        val instance by lazy { AuthApiClient() }
    }
}